# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from yay.nodes import *
from yay.context import Context


class TestSequenceOperations(unittest.TestCase):

    def setUp(self):
        self.ctx = Context(None, None)

    def test_sequence(self):
        l = Sequence([Boxed(1), Boxed(2), Boxed(3)])

        self.failUnlessEqual(l.resolve(self.ctx), [1, 2, 3])

    def test_sequence_append(self):
        l = Sequence([Boxed(1), Boxed(2), Boxed(3)])
        a = Append(Sequence([Boxed(4), Boxed(5)]))
        a.chain = l

        self.failUnlessEqual(a.resolve(self.ctx), [1, 2, 3, 4, 5])

    def test_sequence_remove(self):
        l = Sequence([Boxed(1), Boxed(2), Boxed(3)])
        r = Remove(Sequence([Boxed(2)]))
        r.chain = l

        self.failUnlessEqual(r.resolve(self.ctx), [1, 3])


class TestMapping(unittest.TestCase):

    def setUp(self):
        self.ctx = Context(None, None)

    def test_get_set(self):
        d = Mapping(None)
        d.set("foo", Boxed(1))

        self.failUnlessEqual(d.get(None, "foo", None).value, 1)


class TestLookup(unittest.TestCase):

    def setUp(self):
        self.ctx = Context(None, None)

    def test_lookup(self):
        d = Mapping(None)
        d.set("foo", Boxed(1))
        l = Lookup(d, Boxed("foo"))

        self.failUnlessEqual(l.resolve(self.ctx), 1)




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node

class Boxed(Node):
    """
    A Boxed variable is an unmodified and uninteresting value that
    is wrapped simply so we can put it in our graph
    """

    def resolve_string(self, context, value):
        encountered = set()
        previous = None
        while value != previous:
            if value in encountered:
                self.error("Cycle encountered (%s)" % label)
            encountered.add(value)

            previous = value
            value = value.format(context)

        return value

    def resolve(self, context):
        value = self.value

        if not isinstance(value, basestring):
            return value

        value = self.resolve_string(context, value)

        if value.lower() in ("yes", "true", "on"):
            return True
        if value.lower() in ("no", "false", "off"):
            return False

        try:
            return int(value)
        except ValueError:
            pass

        #FIXME: could still be a float or a timestamp

        try:
            return value.encode('ascii')
        except UnicodeEncodeError:
            return value

    def get(self, context, key, default=None):
        return Boxed(self.value.get(key, default))

    def __repr__(self):
        return "Boxed(%s)" % self.value




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes.node import Node
from yay.nodes.boxed import Boxed
from yay.nodes.mapping import Mapping
from yay.nodes.sequence import Sequence
from yay.nodes.append import Append
from yay.nodes.remove import Remove
from yay.nodes.filter import Filter
from yay.nodes.lookup import Lookup
from yay.nodes.copy import Copy
from yay.nodes.foreach import ForEach
from yay.nodes.function import Function
from yay.nodes.select import Select
from yay.nodes.flatten import Flatten
from yay.nodes.expression import *



# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node

class RootContext(object):

    def __init__(self, root):
        self.root = root
        self.functions = {}

        self.functions['range'] = range

    def get(self, key, default=None):
        return self.root.get(self, key, default)

    def call(self, func, args):
        return self.functions[func](*args)

class Context(object):

    def __init__(self, parent, variables):
        self.parent = parent
        self.variables = variables if variables else {}

    def get(self, key, default=None):
        if key in self.variables:
            return self.variables[key]

        if self.parent:
            return self.parent.get(key, default)

        return default

    def call(self, func, args):
        return self.parent.call(func, args)




from yay.nodes import Node
from yay.context import Context

class Filter(Node):

    def __init__(self, container, filter_expression):
        self.container = container
        self.filter_expression = filter_expression

    def semi_resolve(self, context):
        if not hasattr(self.container, "semi_resolve"):
            self.error("Expected sequence, got '%s'" % self.container)

        resolved = self.container.semi_resolve(context)

        filtered = []
        for r in resolved:
            newctx = Context(context, {"@": r})
            if self.filter_expression.resolve(newctx):
                filtered.append(r)

        return filtered

    def resolve(self, context):
        return list(f.resolve(context) for f in self.semi_resolve(context))

    def __repr__(self):
        return "Filter(%s, %s)" % (self.container, self.filter_expression)



# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import os
import random

from yay.config import Config
from yay.openers import MemOpener
from yay.errors import Error


class TestChaosMonkey(unittest.TestCase):

    def test_chaos(self):
        path = os.path.dirname(__file__)
        raw = open(os.path.join(path, "test_chaosmonkey.yay")).read()

        for i in range(1000):
            idx = random.randint(0, len(raw)-1)
            ln = random.randint(1, 100)

            action = random.choice(["delete", "add"])

            if action == "delete":
                copy = self.delete(raw, idx, ln)
            elif action == "add":
                copy = self.add(raw, idx, ln)

            try:
                MemOpener.add("mem://test_chaosmonkey", copy)
                Config().load_uri("mem://test_chaosmonkey")
            except Error, e:
                pass
            except Exception, e:
                print type(e), idx, len, action
                raise

    def add(self, raw, idx, len):
        return raw[:idx] + " " * len + raw[idx:]

    def delete(self, raw, idx, len):
        return raw[:idx] + raw[idx+len:]




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Boxed

class Function(object):

    def __init__(self, function, args):
        self.function = function
        self.args = args

    def resolve(self, context):
        args = [arg.resolve(context) for arg in self.args]
        return context.call(self.function, args)

    def semi_resolve(self, context):
        return [Boxed(x) for x in self.resolve(context)]

    def __repr__(self):
        return "Function(%s)" % ", ".join([self.function] + [str(a) for a in self.args])




import unittest
import doctest

from yay.config import Config
from yay.openers import MemOpener

class TestConfig(Config):

    def file(self, uri, data):
        MemOpener.add(uri, data)

documentation = [
    'extends.rst',
    'actions.rst',
    ]

globs = {
    "config": TestConfig(),
}


class TestDocumentation(unittest.TestCase):

    def __new__(self, test):
        return getattr(self, test)()

    @classmethod
    def test_documentation(cls):
        return doctest.DocFileSuite(
            *documentation,
            globs = globs
        )




# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node


class Comparison(Node):
    def __init__(self, left, right):
        self.left = left
        self.right = right
    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, self.left, self.right)

class And(Comparison):
    def resolve(self, context):
        return self.left.resolve(context) and self.right.resolve(context)

class Or(Comparison):
    def resolve(self, context):
        return self.left.resolve(context) or self.right.resolve(context)

class In(Comparison):
    def resolve(self, context):
        return self.left.resolve(context) in self.right.resolve(context)

class Equal(Comparison):
    def resolve(self, context):
        return self.left.resolve(context) == self.right.resolve(context)

class NotEqual(Comparison):
    def resolve(self, context):
        return self.left.resolve(context) != self.right.resolve(context)

class LessThan(Comparison):
    def resolve(self, context):
        return self.left.resolve(context) < self.right.resolve(context)

class LessThanEqual(Comparison):
    def resolve(self, context):
        return self.left.resolve(context) <= self.right.resolve(context)

class GreaterThan(Comparison):
    def resolve(self, context):
        return self.left.resolve(context) > self.right.resolve(context)

class GreaterThanEqual(Comparison):
    def resolve(self, context):
        return self.left.resolve(context) >= self.right.resolve(context)


class Access(Node):
    def __init__(self, container, access):
        self.container = container
        self.access = access

    def get(self, context, idx, default=None):
        sr = self.semi_resolve(context)
        return sr.get(context, idx, default)

    def semi_resolve(self, context):
        if self.container:
            unresolved = self.container.get(context, self.access)
            if unresolved is None:
                self.error("Container does not have field '%s'" % self.access)
        else:
            unresolved = context.get(self.access)
            if unresolved is None:
                self.error("Context does not have field '%s'" % self.access)

        if hasattr(unresolved, "semi_resolve"):
            return unresolved.semi_resolve(context)

        return unresolved

    def resolve(self, context):
        sr = self.semi_resolve(context)
        return sr.resolve(context)

    #def get(self, context, key, default=None):
    #    self.resolve(context).get(context, key, default)

    def __repr__(self):
        return "Access(%s, %s)" % (self.container, self.access)


class Concatenation(Node):
    def __init__(self, *args):
        self.args = args

    def resolve(self, context):
        return "".join(str(arg.resolve(context)) for arg in self.args)

    def __repr__(self):
        return "Concat(%s)" % ", ".join(str(arg) for arg in self.args)




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yay.nodes import Node, Lookup, Boxed
from yay.context import Context

def flatten(lst):
    for itm in lst:
        if isinstance(itm, list):
            for x in flatten(itm):
               yield x
        else:
            yield itm

class Flatten(object):

    """ Inspired by Ruby's .flatten - flatten nested lists into a single one """

    def __init__(self, value):
        self.value = value

    def resolve(self, context):
        return list(flatten(self.value.resolve(context)))





# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from yay.config import Config

class TestConfigUpdate(unittest.TestCase):

    def test_simple_update(self):
        c = Config()
        c.load("""
            foo: 1
            bar: 2
            baz: 3
            """)

        self.failUnless('foo' in c.get().keys())

    def test_simple_replace(self):
        c = Config()
        c.load("""
            foo: 1
            bar: 2
            baz: 3
            """)
        c.load("""
            foo: 3
            qux: 1
            """)

        self.failUnlessEqual(c.get()['foo'], 3)
        self.failUnlessEqual(c.get()['bar'], 2)
        self.failUnlessEqual(c.get()['qux'], 1)

    def test_nested_map_update(self):
        c = Config()
        c.load("""
            foo:
                foo: 1
                bar: 2
            bar: 2
            """)
        c.load("""
            foo:
                foo: 2
                baz: 3
            baz: 3
            """)

        self.failUnlessEqual(c.get()['foo']['foo'], 2)
        self.failUnlessEqual(c.get()['foo']['bar'], 2)
        self.failUnlessEqual(c.get()['foo']['baz'], 3)
        self.failUnlessEqual(c.get()['baz'], 3)

    def test_list(self):
        c = Config()
        c.load("""
            foo:
                - 1
                - 2
                - 3
            """)

        self.failUnlessEqual(c.get()['foo'], [1, 2, 3])

    def test_list_append(self):
        c = Config()

        c.load("""
            foo:
                - 1
                - 2
                - 3
            foo.append:
                - 4
                - 5
                - 6
            """)
        self.failUnlessEqual(c.get()['foo'], [1,2,3,4,5,6])

    def test_list_remove(self):
        c = Config()
        c.load("""
            foo:
                - 1
                - 2
                - 3
            foo.remove:
                - 1
                - 2
                - 3
                - 5
            """)

        self.failUnlessEqual(c.get()['foo'], [])




# Copyright 2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import doctest
import os
import glob

import yaml

from yay.config import Config
from yay.openers import MemOpener

dogfood_path = os.path.join(os.path.dirname(__file__), "dogfood")


class TestConfig(Config):

    def load_uri(self, uri):
        return super(TestConfig, self).load_uri("file:///" + os.path.join(dogfood_path, uri))


class TestDogfood(unittest.TestCase):

    class __metaclass__(type):
        def __new__(cls, name, bases, attrs):
            # Generate a testcase for each .in file in the dogfood directory
            for snack in list(os.path.splitext(os.path.basename(x))[0] for x in glob.glob(os.path.join(dogfood_path, "*.in"))):
                name, func = cls.get_test_method(snack)
                attrs[name] = func
            return type.__new__(cls, name, bases, attrs)

        @classmethod
        def get_test_method(cls, snack):
            name = "test_%s" % snack
            load_from = os.path.join(dogfood_path, snack+".in")
            compare_to = os.path.join(dogfood_path, snack+".out")
            def _(self):
               self.failUnlessEqual(self.load(load_from), yaml.load(open(compare_to)))
            _.__name__ = name
            return name, _

    def load(self, path):
        oldpath = os.getcwd()
        os.chdir(dogfood_path)
        try:
            c = TestConfig()
            self.failUnlessEqual(c.get(), {})
            c.load_uri(path)
            return c.get()
        finally:
            os.chdir(oldpath)

    def failUnlessEqual(self, value1, value2):
        if not isinstance(value1, dict) or not isinstance(value2, dict):
            super(TestDogfood, self).failUnlessEqual(value1, value2)
            return

        k1 = frozenset(value1.keys())
        k2 = frozenset(value2.keys())

        if k1.difference(k2):
            raise KeyError("Dictionary 1 contains keys dictionary 2 does not (%s)" % ", ".join(k1.difference(k2)))

        if k2.difference(k1):
            raise KeyError("Dictionary 2 contains keys dictionary 1 does not (%s)" % ", ".join(k2.difference(k1)))

        for key in k1:
            self.failUnlessEqual(value1[key], value2[key])




import os

version = '0.0.15'

from setuptools import setup, find_packages
setup(
    name='yay',
    description='An extensible config file format',
    long_description = open("README.rst").read() + "\n" + \
                        open("CHANGES").read(),
    version=version,
    url='http://pypi.python.org/pypi/yay',
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: System :: Systems Administration",
    ],
    author='John Carr',
    author_email='john.carr@isotoma.com',
    license="Apache Software License",
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    install_requires = [
        "PyYAML",
    ],
)



# Copyright 2010-2011 Isotoma Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Error(Exception):

    def get_string(self):
        return self.args[0]


class LanguageError(Error):

    def __init__(self, description, file, line, column, snippet=None):
        self.description = description
        self.file = file
        self.line = line
        self.column = column
        self.snippet = snippet

    def get_string(self):
        error = self.description
        error += "\nFile %s, line %d, column %d" % (self.file, self.line, self.column)
        if self.snippet:
            error += "\n%s" % self.snippet
        return error

    def __str__(self):
        return self.get_string()


class SyntaxError(LanguageError):
    pass


class EvaluationError(LanguageError):
    pass



